// -*- Mode: c++ -*-
//
// (C) Copyright 1997 OSR Open Systems Resources, Inc.
// All Rights Reserved
//

#include <ntddk.h>
#include <NTDDDISK.H>
//
// Definitions which should be changed for your file system
//

#define FSD_SERVICE_PATH    L"\\Registry\\Machine\\System\\CurrentControlSet\\Services\\MyFsd"
#define FSD_RECOGNIZER_NAME L"\\FileSystem\\MyFsdRecognizer"

#define DEVICE_LOGICAL_BLOCKSIZE 512  // This is the size of your File Systems sectors on disk.

//
// External, non-confidential but unpublished NT entry points
//

NTSYSAPI
NTSTATUS
NTAPI
ZwLoadDriver(
IN PUNICODE_STRING DriverServiceName
);

NTKERNELAPI
VOID
IoRegisterFileSystem(
    IN OUT PDEVICE_OBJECT DeviceObject
    );

NTKERNELAPI
VOID
IoUnregisterFileSystem(
    IN OUT PDEVICE_OBJECT DeviceObject
    );

//
// Global data
//

static PDRIVER_OBJECT RecognizerDriverObject;
static PDEVICE_OBJECT RecognizerDeviceObject;


//
// Forward reference
//

static VOID Unload(PDRIVER_OBJECT);
static NTSTATUS RecognizerFsControl(PDEVICE_OBJECT, PIRP);
static NTSTATUS RecognizerDetectFileSystem(PIRP Irp);
static NTSTATUS RecognizerIoControl( IN PDEVICE_OBJECT deviceObject,
                                     IN ULONG IoctlCode,
                                     IN PVOID InputBuffer,
                                     IN ULONG InputBufferSize,
                                     OUT PVOID OutputBuffer,
                                     OUT ULONG OutputBufferSize);
static NTSTATUS RecognizerIoControl( IN PDEVICE_OBJECT deviceObject,
                                     IN ULONG IoctlCode,
                                     IN PVOID InputBuffer,
                                     IN ULONG InputBufferSize,
                                     OUT PVOID OutputBuffer,
                                     OUT ULONG OutputBufferSize);




//
// DriverEntry
//
//  This is the entry point for the driver.
//
// Inputs:
//  DriverObject - the driver object for this driver
//  RegistryPath - the path to this driver's key
//
// Outputs:
//  None.
//
// Returns:
//  success
//
// Notes:
//  This is a trivial driver, with a trivial entry point.
//
NTSTATUS
DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath)
{
    NTSTATUS code;
    UNICODE_STRING driverName;

    //
    // Save away the driver object pointer
    //

    RecognizerDriverObject = DriverObject;


    //
    // Device object was created.  Set up dispatch entry point for file system control.
    //

    DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] = RecognizerFsControl;


    //
    // Driver itself is unloadable
    //

    DriverObject->DriverUnload = Unload;


    //
    // Initialize the name string for the file system recognizer device object.
    //

    RtlInitUnicodeString(&driverName, FSD_RECOGNIZER_NAME);


    //
    // Create the named device object.
    //

    code = IoCreateDevice(RecognizerDriverObject,
                          0,
                          &driverName,
                          FILE_DEVICE_DISK_FILE_SYSTEM,
                          0,
                          FALSE,
                          &RecognizerDeviceObject);

    if (!NT_SUCCESS(code)) {

        DbgPrint("Recognizer failed to load, failure in IoCreateDevice call returned 0x%x\n", code);

        //
        // Bail out.
        //

        return (code);

    }

    //
    // Register the device object as a file system.
    //

    IoRegisterFileSystem(RecognizerDeviceObject);

    //
    // Done!
    //

    return STATUS_SUCCESS;

}



//
// Unload
//
//  This is the function is called when the OS has determined that their are no more
//  references to the device Object.  It is our job to unregister ourselves as a 
//  File System and delete the Device Object we created.  Typically this will be 
//  called after we have recognized our file systems and told the IO manager to load
//  our file system and get rid of us.
//
// Inputs:
//   DeviceObject - presumably OUR device object
//
// Outputs:
//  None.
//
// Returns:
//  None.
// 
// Notes:
//  None.
// 
static VOID Unload(PDRIVER_OBJECT DriverObject)
{
    //
    // Delete our device object, if there is one, and that's about it.
    //

    if (RecognizerDeviceObject) {

        IoUnregisterFileSystem(RecognizerDeviceObject);

        IoDeleteDevice(RecognizerDeviceObject);

        RecognizerDeviceObject = 0;

    }

    //
    // Done!
    //
}



//
// RecognizerFsControl
//
//  This is the ONLY function that is implemented in this driver and
//  it does two things: handles mount and load requests.  If the mount
//  request is intercepted, it looks at the media.  If the media is
//  recognized, this routine should return LOAD_DRIVER.  When the call
//  comes in from the I/O Manager to load the FSD, it does it.  Simple!
//
// Inputs:
//   DeviceObject - presumably OUR device object
//   Irp - the mount/load request
//
// Outputs:
//  None.
//
// Returns:
//  STATUS_NOT_IMPLEMENTED - something other than mount/load
//  STATUS_UNRECOGNIZED_VOLUME - not ours
//  STATUS_FS_DRIVER_REQUIRED - ours, load us
//  SUCCESS - load was OK
// 
// Notes:
//  None.
// 
static NTSTATUS RecognizerFsControl(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    PIO_STACK_LOCATION irpSp = IoGetCurrentIrpStackLocation(Irp);
    UNICODE_STRING driverName;
    NTSTATUS code;

    //
    // Paranoid, right?
    //

    if (DeviceObject != RecognizerDeviceObject) {
        
        //
        // Not ours.
        //

        Irp->IoStatus.Status = STATUS_NOT_IMPLEMENTED;

        Irp->IoStatus.Information = 0;

        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        return STATUS_NOT_IMPLEMENTED;

    }

    //
    // Verify this is mount/load.  If not, we don't do this.
    //

    if ((irpSp->MinorFunction != IRP_MN_MOUNT_VOLUME) &&
        (irpSp->MinorFunction != IRP_MN_LOAD_FILE_SYSTEM)) {

        //
        // We don't do that (whatever "that" is!)
        //

        Irp->IoStatus.Status = STATUS_NOT_IMPLEMENTED;

        Irp->IoStatus.Information = 0;

        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        return STATUS_NOT_IMPLEMENTED;

    }

    //
    // Handle load case
    //

    if (irpSp->MinorFunction == IRP_MN_LOAD_FILE_SYSTEM) {


       	//
       	// Now, load the file system itself
       	//
            
       	RtlInitUnicodeString(&driverName, FSD_SERVICE_PATH);
            
       	code = ZwLoadDriver(&driverName);

        //
        // Irp completes with the results of the load(s)
        //

        Irp->IoStatus.Status = code;

        Irp->IoStatus.Information = 0;

        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        //
        // In the case of a successful load, we no longer need our device
        // object.
        //

        if (NT_SUCCESS(code)) {

            IoUnregisterFileSystem(RecognizerDeviceObject);

            IoDeleteDevice(RecognizerDeviceObject);

            RecognizerDeviceObject = 0;

            //
            // We're ready to be unloaded!
            //

        }

        return (code);

    }

    //
    // Handle mount case - that's all that's left at this point.
    //

    //
    // XXX: ADD CODE in the RecognizerDetectFileSystem routine TO RECOGNIZE 
    //      THE VOLUME.  Use the routines RecognizerXXXXX routines included 
    //      to issue I/O control requests to the underlying media device.
    //
    code = RecognizerDetectFileSystem(Irp);

    // 
    // Look at the return code.  If this is success, then this is our file system,
    // and we will tell the caller that they should call us back to load our file
    // system drivers.
    //
    if (NT_SUCCESS(code)) {

        //
        // This is the case where the volume is recognized
        //

        Irp->IoStatus.Status = STATUS_FS_DRIVER_REQUIRED;

        Irp->IoStatus.Information = 0;

        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        return STATUS_FS_DRIVER_REQUIRED;
        
    }

    //
    // Volume was not recognized.
    //

    Irp->IoStatus.Status = STATUS_UNRECOGNIZED_VOLUME;
    
    Irp->IoStatus.Information = 0;
    
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    
    return STATUS_UNRECOGNIZED_VOLUME;

}



//
// RecognizerDetectFileSystem
//
//  This is the routine which the user would modify to detect the presence of their
//  File System on the underlying media.
//
// Inputs:
//  Irp - the Irp that requested that the volume be mounted.
//
// Ouputs:
//  None.
//
// Returns:
//  STATUS_SUCCESS - This is our file system and will indicated to the caller to that
//                   our real file system driver needs to be loaded.
//  Other - I/O error
//
// Notes:
//  None.
//
static NTSTATUS RecognizerDetectFileSystem(PIRP Irp)
{
	NTSTATUS								code = STATUS_SUCCESS;
	DISK_GEOMETRY						diskGeometry;
	PARTITION_INFORMATION		partitionInfo;
  PIO_STACK_LOCATION			irpSp = IoGetCurrentIrpStackLocation(Irp);
  PVPB										vpb = irpSp->Parameters.MountVolume.Vpb;
  PDEVICE_OBJECT					mediaDeviceObject = irpSp->Parameters.MountVolume.DeviceObject;
	unsigned char*					pBuffer = NULL;

	//	First build IRP and get Geometry and Partition information for this volume.
  //        use routine RecognizerIoControl to issue these requests
	//	IF error
	//		return error
	//	else
	//		Read the volume master block. Use RecognizerReadDiskSector to issue this request.
	//		If read error
	//			return error
	//		else
	//			unpack the master block  and check signature.
	//			if signature check fails (i.e. this is not your file system)
	//				return error
	//			else
	//				return code STATUS_SUCCESS;
	//			endif
	//		endif
	//	endif
	//

  // For now return an error, but if successful return STATUS_SUCCESS
  code =  STATUS_UNRECOGNIZED_VOLUME;

  return code;
}



//
// RecognizerIoControl
//
//  This routine passes the specified I/O control requested to the underlying device.
//
// Inputs:
//  MediaHandle - the volume handle for the specified device
//  Offset - the logical offset where read starts
//  Length - the length of the read
//  MDL - the mdl chain describing the buffer where the data itself is to be copied
//
// Ouputs:
//  None.
//
// Returns:
//  STATUS_SUCCESS - I/O completed successfully
//  Other - I/O error
//
// Notes:
//  None.
//
static NTSTATUS 
RecognizerIoControl( IN PDEVICE_OBJECT deviceObject,
                     IN ULONG IoctlCode,
                     IN PVOID InputBuffer,
                     IN ULONG InputBufferSize,
                     OUT PVOID OutputBuffer,
                     OUT ULONG OutputBufferSize)
{
    PIRP irp;
    NTSTATUS code;
    KEVENT event;
    IO_STATUS_BLOCK iosb;

    //
    // Init the event we'll use for waiting for the request to complete
    //
    KeInitializeEvent(&event, SynchronizationEvent, FALSE);

    //
    // Build the request
    //
    irp = IoBuildDeviceIoControlRequest(IoctlCode,
                                        deviceObject,
                                        InputBuffer,
                                        InputBufferSize,
                                        OutputBuffer,
                                        OutputBufferSize,
                                        FALSE,
                                        &event,
                                        &iosb);

    //
    // Send the request to the lower layer driver
    //
    code = IoCallDriver(deviceObject, irp);

    //
    // Wait if we must.  Note that we do not accept APCs - we cannot
    // return out of this path until the I/O is done, since the event
    // and IOSB are on the stack.
    //

    if (code == STATUS_PENDING) {

        (void) KeWaitForSingleObject(&event, Executive, KernelMode, TRUE, 0);

        code = iosb.Status;

    }

    //
    // Set the final size of the output buffer.
    //

    OutputBufferSize = iosb.Information;

    //
    // Done!
    //

    return(code);
}



//
//
//	RecognizerReadDiskSector
//
//    Reads a disk sector off the specified input device.  
//
//	Inputs:
//    pDeviceObject - Supplies a pointer to the disk device object.
//		DiskSector - number of the sector on the disk to be read.
//
//	Outputs:
//		Buffer - address of the buffer to receive the read sector.  It must be
//						 sectorSize bytes.
//
//	Returns:
//    Returns TRUE if the disk sector was read.
//
//	Notes:
//		None.
//
static BOOLEAN
RecognizerReadDiskSector(
    IN PDEVICE_OBJECT pDeviceObject,
		IN ULONG	DiskSector,
		IN UCHAR*	Buffer  // must be DEVICE_LOGICAL_BLOCKSIZE bytes long.
    )
{
    LARGE_INTEGER   sectorNumber;
    PIRP            irp;
    IO_STATUS_BLOCK ioStatus;
    KEVENT          event;
    NTSTATUS        status;
    ULONG           sectorSize;
    PULONG          mbr;

    PAGED_CODE();
    sectorNumber.QuadPart = (LONGLONG) DiskSector * DEVICE_LOGICAL_BLOCKSIZE;

    //
    // Create notification event object to be used to signal the inquiry
    // request completion.
    //

    KeInitializeEvent(&event, NotificationEvent, FALSE);

    //
    // Get sector size.
    //

    sectorSize = DEVICE_LOGICAL_BLOCKSIZE;

    //
    // Allocate buffer for sector read.
    //

    mbr = ExAllocatePool(NonPagedPoolCacheAligned, sectorSize);

    if (!mbr) {
        return FALSE;
    }

    //
    // Build IRP to read MBR.
    //

    irp = IoBuildSynchronousFsdRequest(IRP_MJ_READ,
                                       pDeviceObject,
                                       mbr,
                                       sectorSize,
                                       &sectorNumber,
                                       &event,
                                       &ioStatus );

    if (!irp) {
        ExFreePool(mbr);
        return FALSE;
    }

    //
    // Pass request to port driver and wait for request to complete.
    //

    status = IoCallDriver(pDeviceObject,
                          irp);

    if (status == STATUS_PENDING) {
        KeWaitForSingleObject(&event,
                              Suspended,
                              KernelMode,
                              FALSE,
                              NULL);
        status = ioStatus.Status;
    }

    if (!NT_SUCCESS(status)) {
        ExFreePool(mbr);
        return FALSE;
    }

    //
    // return the read partition information
    //
		RtlCopyMemory(Buffer,mbr,sectorSize);

    ExFreePool(mbr);
    return TRUE;
}


